# TimeTable Zope product
#
# Peter Bengtsson <mail@peterbe.com>
# Start date: 03/10 -01
# Last modified date: 21/10 -01
#
# CHANGELOG:
#   - 0.0.2a    Finished SMS stuff and added Account Manager class
#   - 0.0.2     Added primitive SMS transmission capabilities
#               and tucked away stuff into TimeTableUtils instead.
#   - 0.0.1     Built from scratch more or less

__version__='0.0.2a'
__doc__="TimeTable simple object to remember users weekly timetables\
wrapped by getBySMS() and the TimeTableTag."

# python
import string

# Zope
from Products.PageTemplates.PageTemplateFile import PageTemplateFile
from Globals import MessageDialog, Persistent, DTMLFile, default__class_init__
from OFS import SimpleItem, ObjectManager, Folder
from ImageFile import ImageFile
from DocumentTemplate.sequence import SortEx

# This product
from TimeTableUtils import  findEventId, cookDay, cookTime, validTime,\
                            nowsTime, manylists2one, timeAsDateObject,\
                            todaysDay, filterDuplicatesLists, encodeSMSMessage,\
                            wordInList


# Constants
WEEKDAYS =['monday','tuesday','wednesday','thursday','friday','saturday','sunday']
TEXTMESSAGE_MAX_LENGTH=160



#############################################
#
# TIMETABLE STUFF
#


manage_addTimeTableForm = PageTemplateFile('zpt/addTimeTableform', globals())

def manage_addTimeTable(self,id, title="TimeTable",
                      REQUEST=None):
    """ add TimeTable object """
    self._setObject(id, TimeTable(id, title))
    if REQUEST is not None:
        return MessageDialog(title='TimeTable object Added',
              message='This object now has a TimeTable object', 
              action=REQUEST['URL1']+'/manage_main')


class TimeTable(ObjectManager.ObjectManager,
              SimpleItem.SimpleItem,
              Persistent, Folder.Folder):
    """ TimeTable class """
    
    meta_type='TimeTable'

    __ac_roles__ = ('TimeTable User', 'TimeTable Manager')

    
    manage_options = Folder.Folder.manage_options#(
#        {'label': 'Properties','action':'manage_netsend'},
#        {'label': 'Contents', 'action': 'manage_main'},
#        {'label': 'View', 'action': 'index_html'}
#        )
    meta_types = (
        {'name': 'TimeTable Account', 'action': 'manage_addTimeTableAccountForm'},
        {'name': 'TimeTable Manager Account', 'action': 'manage_addTimeTableManagerAccountForm'},
        )
    
    def __init__(self,id,title):
        " "
        self.id = id
        self.title = title
        self.manage_addDTMLDocument('timetable.pdf','TimeTable PDF',\
                                    DTMLFile('dtml/timetable_pdf',globals()))
#        self.colormap = ImageFile('www/colormap.gif',globals())



    def addTimeTableAccount(self,username,password1,password2,
                            telnumber=None, fullname=None,
                            REQUEST=None):
        """ add TimeTable Account object """
        
        if not password1==password2:
            raise "Password Mismatch", "The two passwords are not the same"
        elif not self._validAccountPassword(password1):
            raise "Invalid password", "See guidelines for password"
        
        if hasattr(self, username):
            raise "TakenUsername", "That username is already taken"
        self._setObject(username, TimeTableAccount(username,password1, telnumber, fullname))

        if REQUEST is not None:
            account_object_url = getattr(self,username).absolute_url()
            return MessageDialog(title='TimeTable Account Added',
                                 message='Account URL: <a href="%s">%s</a>'%(account_object_url,account_object_url),
                                 action=REQUEST['URL1']+'/manage_main')

    manage_addTimeTableAccount = addTimeTableAccount

    manage_addTimeTableAccountForm = \
                    PageTemplateFile('zpt/addTimeTableAccountform', globals())

    def addTimeTableManagerAccount(self,username,password1,password2,
                                   telnumber=None, fullname=None,
                                   REQUEST=None):
        """ add TimeTable Manager Account object """
        
        if not password1==password2:
            raise "Password Mismatch", "The two passwords are not the same"
        elif not self._validAccountPassword(password1):
            raise "Invalid password", "See guide    lines for password"
        
        if hasattr(self, username):
            raise "TakenUsername", "That username is already taken"
        self._setObject(username, TimeTableManagerAccount(username,password1, telnumber, fullname))

        if REQUEST is not None:
            account_object_url = getattr(self,username).absolute_url()
            return MessageDialog(title='TimeTable Account Added',
                                 message='Manager Account URL: <a href="%s">%s</a>'%(account_object_url,account_object_url),
                                 action=REQUEST['URL1']+'/manage_main')

    manage_addTimeTableManagerAccount = addTimeTableManagerAccount

    manage_addTimeTableManagerAccountForm = \
                    PageTemplateFile('zpt/addTimeTableManagerAccountform', globals())

    def _validAccountPassword(self,p):
        """ apply policy on password """
        if len(p) <= 2:
            return 0

        return 1
    
    def hasReportLab(self):
        """ test if we can import reportlab """
        try:
            import reportlab
            return 1
        except ImportError:
            return 0

    def getBySMS(self,tel,msg):
        """
            This is an SMS request.
             - Test that the queried account exists
             - Test the account
             - Parse the request
             - return the reply

             To save money, we do NOT reply when things
             go hayward. Only that in debug mode.
        """

        debug = 1 # will always respond

        # Expect first word to be the account
        msg_splitted = string.split(msg)

        if wordInList('help',msg_splitted):
            encoded_msg = encodeSMSMessage("To get your timetable sent to you, \
            let the first word in your message be your accountnumber. \
            This can optionally be followed by a specific day or a time span.")
            return self._splitUpSMS(tel,encoded_msg)

        accountstr = msg_splitted[0].lower()

        requestlist = msg_splitted[1:]

        
        if hasattr(self,accountstr):
            account = getattr(self,accountstr)
            if not account.anonymous_access_sms and \
               account.telnumber != tel:
                if debug:
                    return "%s,sms access not allowed for this user"%tel
                else:
                    return ""

            action = 0  # default, return the rest of the day as long
                        # as possible.
            action_days=[]
            times=[]
            for msg in requestlist:
                
                try:
                    # msg might be 'FRi' and this must be converted to
                    # 'friday', and the try statement if msg='10' or something
                    # stupid that cookDay() won't like.
                    msg = cookDay(msg)
                    if msg in WEEKDAYS:
                        action_days.append(msg)
                except:
                    pass

                try:
                    time = cookTime(msg)
                    if validTime(time):
                        times.append(time)
                except:
                    pass

            if len(times)==0:
                # no specific time asked for, take from now till 23:59

                # Now, if the day asked for isn't "now", the user will
                # much likely know about the whole day that other day.
                # Suppose I on a thursday send "acountnr friday" I will
                # want to know from 00:00 on friday.
                # However, if I on a thursday send "acountnr thursday"
                # which is surpofolous, I will want it from now on.

                if len(action_days)==0 or action_days[0]==todaysDay():
                    times.append(nowsTime())
                else:
                    times.append('00:00')

            if len(times)<2:
                # no end time for the days?, add till the end
                times.append('23:59')

            if len(action_days)==0:
                # no specific day asked for, take today then
                action_days.append(todaysDay())
                
            # go and get the event object for these days
            allevents = account.getEvents()
            allevents = manylists2one(allevents)
            
            # filter them by day
            fresh_events=[]
            for event in allevents:
                if event.day in action_days:
                    fresh_events.append(event)
            # filter them by time
            starttime = times[0]
            endtime = times[1]
            nfresh_events=[]
            tdo = timeAsDateObject
            
            for event in fresh_events:
                if tdo(event.end) > tdo(starttime) and\
                   tdo(event.start) < tdo(endtime):
                    nfresh_events.append(event)
            del fresh_events

            if len(nfresh_events) >0:
                msg_text = self._SMSprintEvents(nfresh_events)
                
                ret_msg = self._splitUpSMS(tel,msg_text)
                

                return ret_msg
            else:
                return "%s,No events for this request"%(tel)
            
                
        else:
            if debug:
                return "%s,no account called that"%tel
            else:
                return ""

    def _splitUpSMS(self,tel,msg):
        """ split the text into several lines.
            Note that this message goes to one user only,
            so the tel should be the same on all lines.
        """
        lines=[]
        while len(msg)> TEXTMESSAGE_MAX_LENGTH or len(lines)==0:
            
            lines.append(msg[:TEXTMESSAGE_MAX_LENGTH])
            
            msg = msg[TEXTMESSAGE_MAX_LENGTH:]

        lines_with_tel=[]
        for line in lines:
            lines_with_tel.append('%s,%s'%(tel, line))

        return string.join(lines_with_tel,'\n')

    def _SMSprintEvents(self,events):
        """ return a SMS suitable string for display """
        strevents =[]
        for event in events:
            s=''
            title = event.title or '(no title)'
            if not title[-1]=='.':
                title=title+'.'
            s=s+'%s-%s '%(event.start,event.end) + title
            if event.kind !='' and event.kind is not None:
                s=s+' kind:%s'%(event.kind)
            if event.description !='' and event.description is not None and\
               event.description != 'None':
                s=s+' (%s)'%(event.description)
                
            strevents.append(s)
            
        whole_message = string.join(strevents,'\\\\n')
        
        return encodeSMSMessage(whole_message)




#############################################
#
# ACCOUNT STUFF
#        
    
class TimeTableAccount(ObjectManager.ObjectManager,
                       SimpleItem.SimpleItem,
                       Persistent, Folder.Folder):
    """ Account using the TimeTable """
    
    meta_type='TimeTable Account'
    icon='misc_/TimeTable/account.gif'

    manage_options = Folder.Folder.manage_options

    meta_types = (
        {'name': 'TimeTable Event', 'action': 'manage_addTimeTableEventForm'},
        )

        
    def __init__(self,username,password,telnumber=None, fullname=None,
                 roles=['TimeTable User']):
        " "
        self.id = username
        self.telnumber = telnumber
        self.fullname = fullname
        self.defaultbgcolor = '#FFF0E1'
        self.anonymous_access_sms = 0
        # create ACL User here
        self.manage_addUserFolder()
        self.timetable_roles = roles
        self.acl_users._doAddUser(username, password, roles,domains='')


    def _RolesforView(self):
        """ which roles should get the View permission """
        return ['Manager','TimeTable User']

    def manage_afterAdd(self,item, container):
        """ now that the object exists, change the security settings """
        item.manage_permission('View',self._RolesforView(),0)


    def allows_anonymous_access(self):
        """
            See if the 'View' permission is aquired
            or not.
        """
        if str(self.acquiredRolesAreUsedBy('View')).lower().strip()=='checked':
            return 1
        
        
        return 0

    def manage_account_action(self):
        """
            look at stuff in the request to
            finally use editAccount()
        """
        req = self.REQUEST

        fullname = req.fullname or None
        telnumber = req.telnumber or None
        defaultbgcolor = req.defaultbgcolor or None
        anonymous_access = req.has_key('anonymous_access') or 0
        anonymous_access_sms = req.has_key('anonymous_access_sms') or 0

        self.editAccount(fullname, telnumber, defaultbgcolor,
                         anonymous_access, anonymous_access_sms)
        return "Account edited"
        
            
    def editAccount(self, fullname=None, telnumber=None,
                    defaultbgcolor='default',
                    anonymous_access=0, anonymous_access_sms=0,
                    REQUEST=None):
        """ change account details """
        if fullname is not None:
            self.fullname = fullname
        if telnumber is not None:
            self.telnumber = telnumber
        if defaultbgcolor is not None:
            self.defaultbgcolor = defaultbgcolor
        self.anonymous_access_sms = anonymous_access_sms

        if anonymous_access:
            # Let anybody view
            self._manage_permission(1)
        else:
            # remove aquired security permission
            self._manage_permission(0)
            
        if REQUEST is not None:
            return MessageDialog(title='TimeTable Account Edited',
                  message='This Account has now been edited', 
                  action=REQUEST['URL1']+'/manage_main')

    def _manage_permission(self, acquire=1):
        """
            Set the view permission on or off
        """
        self.manage_permission('View',self._RolesforView(),acquire=acquire)
        
        
    def addEvent(self, starttime, endtime, day, title,
                 kind='meeting', description='', bgcolor='default',
                 REQUEST=None):
        """ lets you add an event object to an account """

        eventid = findEventId(self)
        self._setObject(eventid, TimeTableEvent(eventid,starttime, endtime, day, title,
                                 kind, description, bgcolor,
                                 REQUEST))
        if REQUEST is not None:
            return MessageDialog(title='TimeTable Event Added',
                  message='This Account now has a new TimeTable Event', 
                  action=REQUEST['URL1']+'/manage_main')
        return eventid

    

    def removeEvent(self, eventid,
                    REQUEST=None):
        """ remove an event from the account """
        self.manage_delObjects([eventid])


    def getHeaders(self):
        """ return list of headers """
        if self.REQUEST.has_key('showDay'):
            return [self.REQUEST['showDay'].lower().capitalize()]
        caps=[]
        for day in WEEKDAYS:
            caps.append('<a href="?showDay=%s">%s</a>'%(day,day.capitalize()))
        return caps

    def getManagerEvents(self,account):
        """ from the account, look in the aq_base for Managers
            events
        """
        events=[]
        managers = account.aq_parent.objectValues('TimeTable ManagerAccount')
        for manager in managers:
            events = events + manager.objectValues('TimeTable Event')
        return events
            
    def getEvents(self):
        """ return list of event objects in list of 7"""
        d={}
        allevents = self.objectValues('TimeTable Event')
        manager_events = self.getManagerEvents(self)
        allevents = allevents + filterDuplicatesLists(allevents, manager_events)
        return self._sortEvents(allevents)

    def _sortEvents(self,events):
        """ arrange events by day """
        if self.REQUEST.has_key('showDay'):
            dayevents=[]
            for event in events:
                if event.day == self.REQUEST['showDay']:
                    dayevents.append(event)
            return [dayevents]
        
        week=[]
        for day in WEEKDAYS:
            exec "%s = []"%day
        for event in events:
            exec "%s.append(event)"%event.day
        for day in WEEKDAYS:
            exec "week.append(%s)"%day

        for day in week:
            print "1st Time"
            for event in day:
                print "Event:", event.title
            day = self._sortDayEvents(day)
            print "2nd Time"
            for event in day:
                print "Event:", event.title

        return week


    def _sortDayEvents(self,day):
        """
            day is a list of events, that needs to be sorted
            by their start property
        """

        for event in day:
            event.updateStartdate()

##        for event in SortEx(day, (('title',),)):
##            print event.startdate, event.title, event.start
        return SortEx(day, (('title',),))
        
    
    def manage_event_action(self):
        """ sniff edit/add and check for data """
        req = self.REQUEST
        
        if req.has_key('removeevent') and req.has_key('eventid'):
            self.manage_removeEvent(req.eventid)
            return "Event removed. Click Save again to restore it"
        elif req.has_key('submitevent') and req.has_key('eventid'):
            # edit an existing event!
            event_o = getattr(self, req.eventid)
            start = req.starttime or None
            end = req.endtime or None
            day = req.day or None
            title = req.title or None
            kind = req.kind or None
            description = req.description or None
            bgcolor = req.bgcolor or 'default'
            event_o.editEvent(start, end, day, title,
                              kind, description, bgcolor)

            return "Event edited"
        elif req.has_key('submitevent'):
            # add a new event
            starttime = req.starttime or None
            endtime = req.endtime or None
            day = req.day or None
            title = req.title or None
            kind = req.kind or None
            description = req.description or None
            bgcolor = req.bgcolor or self.defaultbgcolor
            self.addEvent(starttime, endtime, day, title,
                          kind, description, bgcolor)

            return "Event added"
        else:
            return "Nothing done to event"

    def fullname_or_id(self):
        """ if self.fullname is empty return the id """
        if hasattr(self, 'fullname') and self.fullname !='':
            return self.fullname
        else:
            return self.id
            

    # Aliases    
    index_html = PageTemplateFile('zpt/index_account', globals())
    manage_removeEvent = removeEvent
    manage_addEvent = addEvent
    manage_addTimeTableEventForm = \
                    PageTemplateFile('zpt/addTimeTableEventform', globals())
    display_timetableHTML = DTMLFile('dtml/displayHTML', globals())
    account_settings = PageTemplateFile('zpt/settings_account', globals())
    manage_event = PageTemplateFile('zpt/manage_event', globals())
    manage_editAccount = editAccount
   
        
default__class_init__(TimeTableAccount)




#############################################
#
# MANAGERACCOUNT STUFF
#

class TimeTableManagerAccount(TimeTableAccount):
    """ subclass of normal account. this one is a bit special
        because the events of each Manager is visible for the
        other timetable users
    """
    meta_type='TimeTable ManagerAccount'
    icon='misc_/TimeTable/manager_account.gif'

    def __init__(self,username,password,telnumber=None, fullname=None):
        " "
        TimeTableAccount.__init__(self,username,password,fullname, roles=['TimeTable Manager'])


    def _RolesforView(self):
        """ which roles should get the View permission """
        return ['Manager','TimeTable Manager']
    

       
#############################################
#
# EVENT STUFF
#

class TimeTableEvent(ObjectManager.ObjectManager,
              SimpleItem.SimpleItem,
              Persistent):
    """ events or meetings """

    meta_type='TimeTable Event'
    icon = 'misc_/TimeTable/event.gif'
    
    manage_options = (
        {'label': 'Contents', 'action': 'manage_main'},
        {'label': 'View', 'action': 'index_html'},
        {'label': 'Properties','action':'manage_propertiesForm'}
        )

    def __init__(self,id, start, end, day, title,
                 kind='meeting', description='', bgcolor='default',
                 REQUEST=None):
        """ add event object """
        start = cookTime(start)
        end = cookTime(end)
        day = cookDay(day)
        
        if validTime(start) and validTime(end):
            if self._correctOrder(start,end):
                self.start = start.strip()
                self.end = end.strip()
            else:
                raise "InvalidTime", "Starttime after endtime??"
        else:
            raise "InvalidTimes", "Starttime or endtime \
            invalid. Check their format"
        if self._validDay(day):
            self.day = day.strip().lower()
        else:
            raise "InvalidDay", "The day value was not correctly recognized"
        
        
        self.id = id
        self.title = title
        self.description = description
        self.kind = kind
        self.bgcolor = bgcolor

        return self.id


    def editEvent(self,start=None, end=None, day=None, title=None,
                  kind=None, description=None, bgcolor=None,
                  REQUEST=None):
        """ change properties if passed """

        if start is not None and end is not None:
            # wants to change both.
            # check validness and order
            
            if validTime(start) and validTime(end)\
               and self._correctOrder(start,end):
                self.start = start
                self.end = end
                
            elif start is not None:
                if validTime(start) and \
                   self._correctOrder(start, self.end):
                    self.start = start
                    
            elif end is not None:
                if validTime(end) and \
                   self._correctOrder(self.start, end):
                    self.end = end
        if day is not None and self._validDay(day):
            self.day = day.lower()
            
        if title is not None:
            self.title = title

        if kind is not None:
            self.kind = kind

        if description is not None:
            self.description = description

        if bgcolor is not None:
            self.bgcolor = bgcolor

    def updateStartdate(self):
        """
            Give the event an updated value
            of the startdate property.
            This is to be used for sorting.
        """
        self.startdate = timeAsDateObject(self.start)
        
    def belongs2Manager(self):
        """ looks at the container to see if that object
            is a 'TimeTable ManagerAccount'
        """
        parent = self.aq_parent
        if parent.meta_type=='TimeTable ManagerAccount':
            return 1
        else:
            return 0
        
    def parentName(self):
        """ look at an event and deduce the username of the
            parent account container
        """
        parent = self.aq_parent
        return parent.fullname_or_id()
    
    def getBGcolor(self):
        """ return the bgcolor of an event or the default """
        if hasattr(self,'bgcolor'):
            if self.bgcolor !='default':
                return self.bgcolor

        return self.defaultbgcolor        

    def _validDay(self, d):
        """ check that it's the kind of day
        format we expect
        """
        if d in WEEKDAYS:
            return 1
        return 0
    
    def _correctOrder(self, before, after):
        """ makes sure before is timewise before after """
        
        from DateTime.DateTime import DateTime
        
        b_date = DateTime("2001/12/13 %s"%before)
        a_date = DateTime("2001/12/13 %s"%after)
        if b_date < a_date:
            return 1
        else:
            return 0
        

            

    
    def __str__(self):
        " "
        return "\
        Title: %s\
        Start: %s\
        End: %s\
        Day: %s\
        Kind: %s\
        Description: %s\
        BGcolor: %s"%(self.title, self.start, self.end,
                      self.day, self.kind, self.description,
                      self.bgcolor)
    
        

        
        

